@@ -10,10 +10,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import copy
 from http import client as http_client
 
-from ironic_lib import metrics_utils
 from oslo_utils import uuidutils
 import pecan
 
@@ -27,6 +25,7 @@ from ironic.api import method
 from ironic.common import args
 from ironic.common import exception
 from ironic.common.i18n import _
+from ironic.common import metrics_utils
 from ironic.common import states as ir_states
 from ironic import objects
 
@@ -50,11 +49,7 @@ PORTGROUP_SCHEMA = {
     'additionalProperties': False,
 }
 
-PORTGROUP_PATCH_SCHEMA = copy.deepcopy(PORTGROUP_SCHEMA)
-# patching /extra/vif_port_id has the side-effect of modifying
-# internal_info values, so include it in the patch schema
-PORTGROUP_PATCH_SCHEMA['properties']['internal_info'] = {
-    'type': ['null', 'object']}
+PORTGROUP_PATCH_SCHEMA = PORTGROUP_SCHEMA
 
 PORTGROUP_VALIDATOR_EXTRA = args.dict_valid(
     address=args.mac_address,
@@ -95,10 +90,10 @@ def convert_with_links(rpc_portgroup, fields=None, sanitize=True):
             'mode',
             'name',
             'properties',
-            'standalone_ports_supported'
+            'standalone_ports_supported',
+            'node_uuid'
         )
     )
-    api_utils.populate_node_uuid(rpc_portgroup, portgroup)
     url = api.request.public_url
     portgroup['ports'] = [
         link.make_link('self', url, 'portgroups',
@@ -118,8 +113,7 @@ def convert_with_links(rpc_portgroup, fields=None, sanitize=True):
     return portgroup
 
 
-def list_convert_with_links(rpc_portgroups, limit, url=None, fields=None,
-                            **kwargs):
+def list_convert_with_links(rpc_portgroups, limit, url, fields=None, **kwargs):
     return collection.list_convert_with_links(
         items=[convert_with_links(p, fields=fields, sanitize=False)
                for p in rpc_portgroups],
@@ -153,6 +147,8 @@ class PortgroupsController(pecan.rest.RestController):
             ident = args.uuid_or_name('portgroup', ident)
         except exception.InvalidParameterValue as e:
             pecan.abort(http_client.BAD_REQUEST, e.args[0])
+        except exception.InvalidUuidOrName as e:
+            pecan.abort(http_client.BAD_REQUEST, e.args[0])
         if not remainder:
             return
         subcontroller = self._subcontroller_map.get(remainder[0])
@@ -170,7 +166,8 @@ class PortgroupsController(pecan.rest.RestController):
     def _get_portgroups_collection(self, node_ident, address,
                                    marker, limit, sort_key, sort_dir,
                                    resource_url=None, fields=None,
-                                   detail=None):
+                                   detail=None, project=None,
+                                   conductor_groups=None):
         """Return portgroups collection.
 
         :param node_ident: UUID or name of a node.
@@ -182,6 +179,8 @@ class PortgroupsController(pecan.rest.RestController):
         :param resource_url: Optional, URL to the portgroup resource.
         :param fields: Optional, a list with a specified set of fields
                        of the resource to be returned.
+        :param project: Optional, project ID to filter the request by.
+        :param conductor_groups: conductor groups, to filter the request by.
         """
         limit = api_utils.validate_limit(limit)
         sort_dir = api_utils.validate_sort_dir(sort_dir)
@@ -198,6 +197,11 @@ class PortgroupsController(pecan.rest.RestController):
 
         node_ident = self.parent_node_ident or node_ident
 
+        if conductor_groups and (node_ident or address):
+            raise exception.Invalid(
+                _("Filtering by conductor_groups and node_ident/address "
+                  "simultaneously is not supported."))
+
         if node_ident:
             # FIXME: Since all we need is the node ID, we can
             #        make this more efficient by only querying
@@ -206,13 +210,16 @@ class PortgroupsController(pecan.rest.RestController):
             node = api_utils.get_rpc_node(node_ident)
             portgroups = objects.Portgroup.list_by_node_id(
                 api.request.context, node.id, limit,
-                marker_obj, sort_key=sort_key, sort_dir=sort_dir)
+                marker_obj, sort_key=sort_key, sort_dir=sort_dir,
+                project=project)
         elif address:
-            portgroups = self._get_portgroups_by_address(address)
+            portgroups = self._get_portgroups_by_address(address,
+                                                         project=project)
         else:
-            portgroups = objects.Portgroup.list(api.request.context, limit,
-                                                marker_obj, sort_key=sort_key,
-                                                sort_dir=sort_dir)
+            portgroups = objects.Portgroup.list(
+                api.request.context, limit,
+                marker_obj, sort_key=sort_key, sort_dir=sort_dir,
+                project=project, conductor_groups=conductor_groups)
         parameters = {}
         if detail is not None:
             parameters['detail'] = detail
@@ -224,7 +231,7 @@ class PortgroupsController(pecan.rest.RestController):
                                        sort_dir=sort_dir,
                                        **parameters)
 
-    def _get_portgroups_by_address(self, address):
+    def _get_portgroups_by_address(self, address, project=None):
         """Retrieve a portgroup by its address.
 
         :param address: MAC address of a portgroup, to get the portgroup
@@ -235,7 +242,8 @@ class PortgroupsController(pecan.rest.RestController):
         """
         try:
             portgroup = objects.Portgroup.get_by_address(api.request.context,
-                                                         address)
+                                                         address,
+                                                         project=project)
             return [portgroup]
         except exception.PortgroupNotFound:
             return []
@@ -245,10 +253,10 @@ class PortgroupsController(pecan.rest.RestController):
     @args.validate(node=args.uuid_or_name, address=args.mac_address,
                    marker=args.uuid, limit=args.integer, sort_key=args.string,
                    sort_dir=args.string, fields=args.string_list,
-                   detail=args.boolean)
+                   detail=args.boolean, conductor_groups=args.string_list)
     def get_all(self, node=None, address=None, marker=None,
                 limit=None, sort_key='id', sort_dir='asc', fields=None,
-                detail=None):
+                detail=None, conductor_groups=None):
         """Retrieve a list of portgroups.
 
         :param node: UUID or name of a node, to get only portgroups for that
@@ -264,11 +272,19 @@ class PortgroupsController(pecan.rest.RestController):
         :param sort_dir: direction to sort. "asc" or "desc". Default: asc.
         :param fields: Optional, a list with a specified set of fields
                        of the resource to be returned.
+        :param conductor_groups: conductor groups, to filter the request by.
         """
         if not api_utils.allow_portgroups():
             raise exception.NotFound()
 
-        api_utils.check_policy('baremetal:portgroup:get')
+        if self.parent_node_ident:
+            # Override the node, since this is being called by another
+            # controller with a linked view.
+            node = self.parent_node_ident
+
+        project = api_utils.check_port_list_policy(
+            portgroup=True,
+            parent_node=self.parent_node_ident)
 
         api_utils.check_allowed_portgroup_fields(fields)
         api_utils.check_allowed_portgroup_fields([sort_key])
@@ -276,19 +292,22 @@ class PortgroupsController(pecan.rest.RestController):
         fields = api_utils.get_request_return_fields(fields, detail,
                                                      _DEFAULT_RETURN_FIELDS)
 
-        return self._get_portgroups_collection(node, address,
-                                               marker, limit,
-                                               sort_key, sort_dir,
-                                               fields=fields,
-                                               detail=detail)
+        return self._get_portgroups_collection(
+            node, address, marker, limit, sort_key, sort_dir,
+            fields=fields,
+            resource_url='portgroups',
+            detail=detail,
+            project=project,
+            conductor_groups=conductor_groups)
 
     @METRICS.timer('PortgroupsController.detail')
     @method.expose()
     @args.validate(node=args.uuid_or_name, address=args.mac_address,
                    marker=args.uuid, limit=args.integer, sort_key=args.string,
-                   sort_dir=args.string)
+                   sort_dir=args.string, conductor_groups=args.string_list)
     def detail(self, node=None, address=None, marker=None,
-               limit=None, sort_key='id', sort_dir='asc'):
+               limit=None, sort_key='id', sort_dir='asc',
+               conductor_groups=None):
         """Retrieve a list of portgroups with detail.
 
         :param node: UUID or name of a node, to get only portgroups for that
@@ -302,11 +321,20 @@ class PortgroupsController(pecan.rest.RestController):
                       max_limit resources will be returned.
         :param sort_key: column to sort results by. Default: id.
         :param sort_dir: direction to sort. "asc" or "desc". Default: asc.
+        :param conductor_groups: conductor groups, to filter the request by.
         """
         if not api_utils.allow_portgroups():
             raise exception.NotFound()
 
-        api_utils.check_policy('baremetal:portgroup:get')
+        if self.parent_node_ident:
+            # If we have a parent node, then we need to override this method's
+            # node filter.
+            node = self.parent_node_ident
+
+        project = api_utils.check_port_list_policy(
+            portgroup=True,
+            parent_node=self.parent_node_ident)
+
         api_utils.check_allowed_portgroup_fields([sort_key])
 
         # NOTE: /detail should only work against collections
@@ -314,10 +342,10 @@ class PortgroupsController(pecan.rest.RestController):
         if parent != "portgroups":
             raise exception.HTTPNotFound()
 
-        resource_url = '/'.join(['portgroups', 'detail'])
         return self._get_portgroups_collection(
             node, address, marker, limit, sort_key, sort_dir,
-            resource_url=resource_url)
+            resource_url='portgroups/detail', project=project,
+            conductor_groups=conductor_groups)
 
     @METRICS.timer('PortgroupsController.get_one')
     @method.expose()
@@ -332,7 +360,8 @@ class PortgroupsController(pecan.rest.RestController):
         if not api_utils.allow_portgroups():
             raise exception.NotFound()
 
-        api_utils.check_policy('baremetal:portgroup:get')
+        rpc_portgroup, rpc_node = api_utils.check_port_policy_and_retrieve(
+            'baremetal:portgroup:get', portgroup_ident, portgroup=True)
 
         if self.parent_node_ident:
             raise exception.OperationNotPermitted()
@@ -355,8 +384,31 @@ class PortgroupsController(pecan.rest.RestController):
         if not api_utils.allow_portgroups():
             raise exception.NotFound()
 
+        raise_node_not_found = False
+        node = None
+        owner = None
+        lessee = None
+        node_uuid = portgroup.get('node_uuid')
+        try:
+            # The replace_node_uuid_with_id also checks access to the node
+            # and will raise an exception if access is not permitted.
+            node = api_utils.replace_node_uuid_with_id(portgroup)
+            owner = node.owner
+            lessee = node.lessee
+        except exception.NotFound:
+            raise_node_not_found = True
+
+        # While the rule is for the port, the base object that controls access
+        # is the node.
+        api_utils.check_owner_policy('node', 'baremetal:portgroup:create',
+                                     owner, lessee=lessee,
+                                     conceal_node=False)
+        if raise_node_not_found:
+            # Delayed raise of NodeNotFound because we want to check
+            # the access policy first.
+            raise exception.NodeNotFound(node=node_uuid,
+                                         code=http_client.BAD_REQUEST)
         context = api.request.context
-        api_utils.check_policy('baremetal:portgroup:create')
 
         if self.parent_node_ident:
             raise exception.OperationNotPermitted()
@@ -372,14 +424,10 @@ class PortgroupsController(pecan.rest.RestController):
             raise exception.ClientSideError(
                 error_msg, status_code=http_client.BAD_REQUEST)
 
-        api_utils.handle_post_port_like_extra_vif(portgroup)
-
         # NOTE(yuriyz): UUID is mandatory for notifications payload
         if not portgroup.get('uuid'):
             portgroup['uuid'] = uuidutils.generate_uuid()
 
-        node = api_utils.replace_node_uuid_with_id(portgroup)
-
         new_portgroup = objects.Portgroup(context, **portgroup)
 
         notify.emit_start_notification(context, new_portgroup, 'create',
@@ -409,7 +457,9 @@ class PortgroupsController(pecan.rest.RestController):
             raise exception.NotFound()
 
         context = api.request.context
-        api_utils.check_policy('baremetal:portgroup:update')
+
+        rpc_portgroup, rpc_node = api_utils.check_port_policy_and_retrieve(
+            'baremetal:portgroup:update', portgroup_ident, portgroup=True)
 
         if self.parent_node_ident:
             raise exception.OperationNotPermitted()
@@ -421,9 +471,6 @@ class PortgroupsController(pecan.rest.RestController):
 
         api_utils.patch_validate_allowed_fields(patch, PATCH_ALLOWED_FIELDS)
 
-        rpc_portgroup = api_utils.get_rpc_portgroup_with_suffix(
-            portgroup_ident)
-
         names = api_utils.get_patch_values(patch, '/name')
         for name in names:
             if (name and not api_utils.is_valid_logical_name(name)):
@@ -440,8 +487,8 @@ class PortgroupsController(pecan.rest.RestController):
         # 1) Remove node_id because it's an internal value and
         #    not present in the API object
         # 2) Add node_uuid
-        rpc_node = api_utils.replace_node_id_with_uuid(portgroup_dict)
-
+        portgroup_dict.pop('node_id')
+        portgroup_dict['node_uuid'] = rpc_node.uuid
         portgroup_dict = api_utils.apply_jsonpatch(portgroup_dict, patch)
 
         if 'mode' not in portgroup_dict:
@@ -459,9 +506,6 @@ class PortgroupsController(pecan.rest.RestController):
             e.code = http_client.BAD_REQUEST  # BadRequest
             raise
 
-        api_utils.handle_patch_port_like_extra_vif(
-            rpc_portgroup, portgroup_dict.get('internal_info'), patch)
-
         api_utils.patched_validate_with_schema(
             portgroup_dict, PORTGROUP_PATCH_SCHEMA, PORTGROUP_PATCH_VALIDATOR)
 
@@ -504,17 +548,14 @@ class PortgroupsController(pecan.rest.RestController):
         if not api_utils.allow_portgroups():
             raise exception.NotFound()
 
+        rpc_portgroup, rpc_node = api_utils.check_port_policy_and_retrieve(
+            'baremetal:portgroup:delete', portgroup_ident, portgroup=True)
+
         context = api.request.context
-        api_utils.check_policy('baremetal:portgroup:delete')
 
         if self.parent_node_ident:
             raise exception.OperationNotPermitted()
 
-        rpc_portgroup = api_utils.get_rpc_portgroup_with_suffix(
-            portgroup_ident)
-        rpc_node = objects.Node.get_by_id(api.request.context,
-                                          rpc_portgroup.node_id)
-
         notify.emit_start_notification(context, rpc_portgroup, 'delete',
                                        node_uuid=rpc_node.uuid)
         with notify.handle_error_notification(context, rpc_portgroup, 'delete',
