@@ -526,7 +526,7 @@ class ServiceUnavailable(IronicException):
 
 
 class Forbidden(IronicException):
-    _msg_fmt = _("Requested OpenStack Images API is forbidden")
+    _msg_fmt = _("Requested OpenStack Baremetal API is forbidden")
 
 
 # TODO(dtantsur): leave only one variant
@@ -1062,3 +1062,87 @@ class ServiceRegistrationFailure(IronicException):
 class Unauthorized(IronicException):
     code = http_client.UNAUTHORIZED
     headers = {'WWW-Authenticate': 'Basic realm="Baremetal API"'}
+
+
+class SecurityProxyNegotiationFailed(IronicException):
+    _msg_fmt = _("Failed to negotiate security type with server: %(reason)s")
+
+
+class RFBAuthHandshakeFailed(IronicException):
+    _msg_fmt = _("Failed to complete auth handshake: %(reason)s")
+
+
+class RFBAuthNoAvailableScheme(IronicException):
+    _msg_fmt = _("No matching auth scheme: allowed types: "
+                 "'%(allowed_types)s', "
+                 "desired types: '%(desired_types)s'")
+
+
+class ConsoleContainerError(IronicException):
+    _msg_fmt = _("Console container error with provider '%(provider)s', "
+                 "reason: %(reason)s")
+
+
+class ImageHostRateLimitFailure(TemporaryFailure):
+    _msg_fmt = _("The image registry has indicates the rate limit has been "
+                 "exceeded for url %(image_ref)s. Please try again later or "
+                 "consider using authentication and/or trying again.")
+
+
+class ImageMatchFailure(InvalidImage):
+    _msg_fmt = _("The requested image lacks the required attributes to "
+                 "identify the file to select.")
+
+
+class OciImageNotSpecific(InvalidImage):
+    _msg_fmt = _("The requested image (%(image_ref)s) was not specific. "
+                 "Please supply a full URL mapping to the manifest to be "
+                 "utilized for the file download.")
+
+
+class OciImageTagNotFound(ImageNotFound):
+    _msg_fmt = _("Image tag %(image_url)s could not be found, "
+                 "available tags are %(tags)s")
+
+
+class ImageServiceAuthenticationRequired(ImageUnacceptable):
+    _msg_fmt = _("The requested image %(image_ref)s requires "
+                 "authentication which has not been provided. "
+                 "Unable to proceed.")
+
+
+class InspectionRuleAlreadyExists(Conflict):
+    """Rule requested already exists in the database."""
+    _msg_fmt = _("A rule with UUID %(uuid)s already exists.")
+
+
+class InspectionRuleNotFound(NotFound):
+    """The requested rule was not found."""
+    _msg_fmt = _("Rule %(rule)s could not be found.")
+
+
+class InspectionRuleValidationFailure(IronicException):
+    """Inspection rule validation fails during creation or execution."""
+    _msg_fmt = _("Inspection rule validation failed. Reason: %(reason)s")
+
+
+class InspectionRuleExecutionFailure(HardwareInspectionFailure):
+    """Raised when an inspection rule fails during execution."""
+    _msg_fmt = _("Inspection rule execution failed. Reason: %(reason)s")
+
+
+class RuleActionExecutionFailure(InspectionRuleExecutionFailure):
+    """Raised when an inspection rule action fails during execution."""
+    _msg_fmt = _("Inspection rule action execution failed. "
+                 "Reason: %(reason)s")
+
+
+class RuleConditionCheckFailure(InspectionRuleExecutionFailure):
+    """Raised when an inspection rule condition fails during execution."""
+    _msg_fmt = _("Inspection rule condition check failed. Reason: %(reason)s")
+
+
+class ConfigDriveRegenerationFailure(IronicException):
+    """Raised when we fail to handle configuration drive corrections."""
+    # NOTE(TheJulia): This is not intended to get raised to a user, but more
+    # so we handle known possible failure cases and don't fail horribly.
