@@ -16,11 +16,11 @@
 
 import copy
 import os
+import pathlib
 import shutil
 import tempfile
 from urllib import parse as urlparse
 
-from ironic_lib import utils as ironic_utils
 import jinja2
 from oslo_concurrency import processutils
 from oslo_log import log as logging
@@ -108,7 +108,7 @@ def _link_mac_pxe_configs(task, ipxe_enabled=False):
     """
 
     def create_link(mac_path):
-        ironic_utils.unlink_without_raise(mac_path)
+        utils.unlink_without_raise(mac_path)
         relative_source_path = os.path.relpath(
             pxe_config_file_path, os.path.dirname(mac_path))
         utils.create_link_without_raise(relative_source_path, mac_path)
@@ -159,7 +159,7 @@ def _link_ip_address_pxe_configs(task, ipxe_enabled=False):
         ip_addrs = []
     for port_ip_address in ip_addrs:
         ip_address_path = _get_pxe_ip_address_path(port_ip_address)
-        ironic_utils.unlink_without_raise(ip_address_path)
+        utils.unlink_without_raise(ip_address_path)
         relative_source_path = os.path.relpath(
             pxe_config_file_path, os.path.dirname(ip_address_path))
         utils.create_link_without_raise(relative_source_path,
@@ -429,18 +429,19 @@ def clean_up_pxe_config(task, ipxe_enabled=False):
             except exception.FailedToGetIPAddressOnPort:
                 continue
             # Cleaning up config files created for grub2.
-            ironic_utils.unlink_without_raise(ip_address_path)
+            utils.unlink_without_raise(ip_address_path)
 
     for port in task.ports:
         client_id = port.extra.get('client-id')
         # syslinux, ipxe, etc.
-        ironic_utils.unlink_without_raise(
-            _get_pxe_mac_path(port.address, client_id=client_id,
+        utils.unlink_without_raise(
+            _get_pxe_mac_path(port.address,
+                              client_id=client_id,
                               ipxe_enabled=ipxe_enabled))
         # Grub2 MAC address based configuration
         for path in _get_pxe_grub_mac_path(port.address,
                                            ipxe_enabled=ipxe_enabled):
-            ironic_utils.unlink_without_raise(path)
+            utils.unlink_without_raise(path)
     utils.rmtree_without_raise(os.path.join(_get_root_dir(ipxe_enabled),
                                             task.node.uuid))
 
@@ -666,20 +667,11 @@ def parse_driver_info(node, mode='deploy'):
         # options. Skipping.
         return {}
 
-    info = node.driver_info
-
-    params_to_check = KERNEL_RAMDISK_LABELS[mode]
-
-    d_info = {k: info.get(k) for k in params_to_check}
-    if not any(d_info.values()):
-        # NOTE(dtantsur): avoid situation when e.g. deploy_kernel comes from
-        # driver_info but deploy_ramdisk comes from configuration, since it's
-        # a sign of a potential operator's mistake.
-        d_info = {k: getattr(CONF.conductor, k) for k in params_to_check}
+    info = driver_utils.get_agent_kernel_ramdisk(node, mode=mode)
     error_msg = _("Cannot validate PXE bootloader. Some parameters were"
                   " missing in node's driver_info and configuration")
-    deploy_utils.check_for_missing_params(d_info, error_msg)
-    return d_info
+    deploy_utils.check_for_missing_params(info, error_msg)
+    return info
 
 
 def get_instance_image_info(task, ipxe_enabled=False):
@@ -1164,7 +1156,7 @@ def validate_kickstart_template(ks_template):
         msg = (_("The kickstart template includes a variable that is not "
                  "a valid kickstart option. Rendering the template returned "
                  " %(msg)s. The valid options are %(valid_options)s.") %
-               {'msg': exc.message,
+               {'msg': exc,
                 'valid_options': ','.join(ks_options.keys())})
         raise exception.InvalidKickstartTemplate(msg)
 
@@ -1328,8 +1320,16 @@ def cache_ramdisk_kernel(task, pxe_info, ipxe_enabled=False):
     LOG.debug("Fetching necessary kernel and ramdisk for node %s",
               node.uuid)
     images_info = list(t_pxe_info.values())
+
+    # Call the central auth override lookup, but signal it is not
+    # for *user* direct artifacts, i.e. this is system related
+    # activity as we're getting TFTP Image cache artifacts.
+    img_service_auth = service.get_image_service_auth_override(
+        task.node, permit_user_auth=False)
+
     deploy_utils.fetch_images(ctx, TFTPImageCache(), images_info,
-                              CONF.force_raw_images)
+                              CONF.force_raw_images,
+                              image_auth_data=img_service_auth)
     if CONF.pxe.file_permission:
         for info in images_info:
             os.chmod(info[1], CONF.pxe.file_permission)
@@ -1348,13 +1348,13 @@ def clean_up_pxe_env(task, images_info, ipxe_enabled=False):
     """
     for label in images_info:
         path = images_info[label][1]
-        ironic_utils.unlink_without_raise(path)
+        utils.unlink_without_raise(path)
 
     clean_up_pxe_config(task, ipxe_enabled=ipxe_enabled)
     TFTPImageCache().clean_up()
 
 
-def place_loaders_for_boot(base_path):
+def place_loaders_for_boot(base_path: str | os.PathLike):
     """Place configured bootloaders from the host OS.
 
     Example: grubaa64.efi:/path/to/grub-aarch64.efi,...
@@ -1365,29 +1365,27 @@ def place_loaders_for_boot(base_path):
     if not loaders or not base_path:
         # Do nothing, return.
         return
+    dir_mode = CONF.pxe.dir_permission or 0o755
+
+    base_path = pathlib.Path(base_path)
 
     for dest, src in loaders.items():
-        (head, _tail) = os.path.split(dest)
-        if head:
-            if head.startswith('/'):
-                # NOTE(TheJulia): The intent here is to error if the operator
-                # has put absolute paths in place, as we can and likely should
-                # copy to multiple folders based upon the protocol operation
-                # being used. Absolute paths are problematic there, where
-                # as a relative path is more a "well, that is silly, but okay
-                # misconfiguration.
-                msg = ('File paths configured for [pxe]loader_file_paths '
-                       'must be relative paths. Entry: %s') % dest
-                raise exception.IncorrectConfiguration(msg)
-            else:
-                try:
-                    ensure_tree(os.path.join(base_path, head))
-                except OSError as e:
-                    msg = ('Failed to create supplied directories in '
-                           'asset copy paths. Error: %s') % e
-                    raise exception.IncorrectConfiguration(msg)
+        dest = pathlib.Path(dest)
+        if dest.is_absolute():
+            msg = ('File paths configured for [pxe]loader_file_paths '
+                   'must be relative paths. Entry: %s') % dest
+            raise exception.IncorrectConfiguration(msg)
+
+        full_dest = base_path.joinpath(dest)
+        try:
+            # ensure that we have somewhere to copy the files to
+            # notice the use of the parent path of the file
+            full_dest.parent.mkdir(mode=dir_mode, parents=True, exist_ok=True)
+        except OSError as e:
+            msg = (_('Failed to create bootloader destination path. Error: %s')
+                   % e)
+            raise exception.IncorrectConfiguration(msg)
 
-        full_dest = os.path.join(base_path, dest)
         LOG.debug('Copying bootloader %(dest)s from %(src)s.',
                   {'src': src, 'dest': full_dest})
         try:
