@@ -53,6 +53,7 @@ VERBS = {
     'rescue': 'rescue',
     'unrescue': 'unrescue',
     'unhold': 'unhold',
+    'service': 'service',
 }
 """ Mapping of state-changing events that are PUT to the REST API
 
@@ -235,11 +236,32 @@ UNRESCUEFAIL = 'unrescue failed'
 UNRESCUING = 'unrescuing'
 """ Node is being restored from rescue mode (to active state). """
 
+SERVICE = 'service'
+""" Node is being requested to be modified through a service step. """
+
+SERVICING = 'servicing'
+""" Node is actively being changed by a service step. """
+
+SERVICEWAIT = 'service wait'
+""" Node is waiting for an operation to complete. """
+
+SERVICEFAIL = 'service failed'
+""" Node has failed in a service step execution. """
+
+SERVICEHOLD = 'service hold'
+""" Node is being held for direct intervention from a service step. """
+
+
+"""All Node states related to servicing."""
+SERVICING_STATES = frozenset((SERVICING, SERVICEWAIT,
+                              SERVICEFAIL, SERVICEHOLD))
+
+
 # NOTE(kaifeng): INSPECTING is allowed to keep backwards compatibility,
 # starting from API 1.39 node update is disallowed in this state.
 UPDATE_ALLOWED_STATES = (DEPLOYFAIL, INSPECTING, INSPECTFAIL, INSPECTWAIT,
                          CLEANFAIL, ERROR, VERIFYING, ADOPTFAIL, RESCUEFAIL,
-                         UNRESCUEFAIL)
+                         UNRESCUEFAIL, SERVICE, SERVICEHOLD, SERVICEFAIL)
 """Transitional states in which we allow updating a node."""
 
 DELETE_ALLOWED_STATES = (MANAGEABLE, ENROLL, ADOPTFAIL)
@@ -250,11 +272,12 @@ STABLE_STATES = (ENROLL, MANAGEABLE, AVAILABLE, ACTIVE, ERROR, RESCUE)
 
 UNSTABLE_STATES = (DEPLOYING, DEPLOYWAIT, CLEANING, CLEANWAIT, VERIFYING,
                    DELETING, INSPECTING, INSPECTWAIT, ADOPTING, RESCUING,
-                   RESCUEWAIT, UNRESCUING)
+                   RESCUEWAIT, UNRESCUING, SERVICING, SERVICEWAIT)
 """States that can be changed without external request."""
 
 STUCK_STATES_TREATED_AS_FAIL = (DEPLOYING, CLEANING, VERIFYING, INSPECTING,
-                                ADOPTING, RESCUING, UNRESCUING, DELETING)
+                                ADOPTING, RESCUING, UNRESCUING, DELETING,
+                                SERVICING)
 """States that cannot be resumed once a conductor dies.
 
 If a node gets stuck with one of these states for some reason
@@ -272,12 +295,15 @@ _FASTTRACK_LOOKUP_ALLOWED_STATES = (ENROLL, MANAGEABLE, AVAILABLE,
                                     DEPLOYING, DEPLOYWAIT,
                                     CLEANING, CLEANWAIT,
                                     INSPECTING, INSPECTWAIT,
-                                    RESCUING, RESCUEWAIT)
+                                    RESCUING, RESCUEWAIT,
+                                    SERVICING, SERVICEWAIT,
+                                    SERVICEHOLD)
 FASTTRACK_LOOKUP_ALLOWED_STATES = frozenset(_FASTTRACK_LOOKUP_ALLOWED_STATES)
 """States where API lookups are permitted with fast track enabled."""
 
 FAILURE_STATES = frozenset((DEPLOYFAIL, CLEANFAIL, INSPECTFAIL,
-                            RESCUEFAIL, UNRESCUEFAIL, ADOPTFAIL))
+                            RESCUEFAIL, UNRESCUEFAIL, ADOPTFAIL,
+                            SERVICEFAIL))
 
 
 ##############
@@ -594,3 +620,55 @@ machine.add_transition(ADOPTFAIL, ADOPTING, 'adopt')
 
 # A node that failed adoption can be moved back to manageable
 machine.add_transition(ADOPTFAIL, MANAGEABLE, 'manage')
+
+# Add service* states
+machine.add_state(SERVICING, target=ACTIVE, **watchers)
+machine.add_state(SERVICEWAIT, target=ACTIVE, **watchers)
+machine.add_state(SERVICEFAIL, target=ACTIVE, **watchers)
+machine.add_state(SERVICEHOLD, target=ACTIVE, **watchers)
+
+# A node in service an be returned to active
+machine.add_transition(SERVICING, ACTIVE, 'done')
+
+# A node in active can be serviced
+machine.add_transition(ACTIVE, SERVICING, 'service')
+
+# A node in servicing can be failed
+machine.add_transition(SERVICING, SERVICEFAIL, 'fail')
+
+# A node in service can enter a wait state
+machine.add_transition(SERVICING, SERVICEWAIT, 'wait')
+
+# A node in service can be held
+machine.add_transition(SERVICING, SERVICEHOLD, 'hold')
+machine.add_transition(SERVICEWAIT, SERVICEHOLD, 'hold')
+
+# A held node in service can get more service steps to start over
+machine.add_transition(SERVICEHOLD, SERVICING, 'service')
+
+# A held node in service can be removed from service
+machine.add_transition(SERVICEHOLD, SERVICEWAIT, 'unhold')
+
+# A node in service wait can resume
+machine.add_transition(SERVICEWAIT, SERVICING, 'resume')
+
+# A node in service wait can failed
+machine.add_transition(SERVICEWAIT, SERVICEFAIL, 'fail')
+
+# A node in service wait can be aborted
+machine.add_transition(SERVICEWAIT, SERVICEFAIL, 'abort')
+
+# A node in service hold can be aborted
+machine.add_transition(SERVICEHOLD, SERVICEFAIL, 'abort')
+
+# A node in service fail can re-enter service
+machine.add_transition(SERVICEFAIL, SERVICING, 'service')
+
+# A node in service fail can be rescued
+machine.add_transition(SERVICEFAIL, RESCUING, 'rescue')
+
+# A node in service fail may be deleted.
+machine.add_transition(SERVICEFAIL, DELETING, 'delete')
+
+# A node in service wait may be deleted.
+machine.add_transition(SERVICEWAIT, DELETING, 'delete')
