@@ -18,6 +18,7 @@
 from oslo_config import cfg
 from oslo_config import types
 
+from ironic.common import automated_clean_methods
 from ironic.common import boot_modes
 from ironic.common.i18n import _
 from ironic.common import lessee_sources
@@ -146,12 +147,71 @@ opts = [
                        'state. When enabled, the particular steps '
                        'performed to clean a node depend on which driver '
                        'that node is managed by; see the individual '
-                       'driver\'s documentation for details. '
-                       'NOTE: The introduction of the cleaning operation '
-                       'causes instance deletion to take significantly '
-                       'longer. In an environment where all tenants are '
-                       'trusted (eg, because there is only one tenant), '
-                       'this option could be safely disabled.')),
+                       'driver\'s documentation for details.')),
+    cfg.StrOpt('automated_cleaning_step_source',
+               default='autogenerated',
+               mutable=True,
+               choices=[
+                   (automated_clean_methods.AUTOGENERATED, _(
+                       'Collects steps from hardware interfaces and orders '
+                       'by priority. This provides the original Ironic '
+                       'cleaning behavior originally implemented in Kilo.')),
+                   (automated_clean_methods.RUNBOOK, _(
+                       'Runs cleaning via a runbook specified in '
+                       'configuration or node driver_info. If a runbook '
+                       'is not specified while automated_clean is enabled, '
+                       'cleaning will fail.')),
+                   (automated_clean_methods.HYBRID, _(
+                       'Runs cleaning via a runbook if one is specified in '
+                       'configuration or node driver_info. If a runbook '
+                       'is not specified while automated_clean is enabled, '
+                       'Ironic will fallback to \'autogenerated\' '
+                       'cleaning steps.'))
+               ],
+               help=_('Determines how automated_cleaning is performed; the '
+                      'default, \'autogenerated\' collects steps from '
+                      'hardware interfaces, then ordering by priority; '
+                      '\'runbook\' requires a runbook to be specified in '
+                      'config or driver_info, which is then used to clean the'
+                      'node; \'hybrid\' uses a runbook if available, '
+                      'and falls-back to autogenerated cleaning steps if not.'
+                      )),
+    cfg.StrOpt('automated_cleaning_runbook',
+               default=None,
+               mutable=True,
+               help=_('If set and [conductor]/automated_clean_step_source '
+                      'is set to \'hybrid\' or \'runbook\', the runbook '
+                      'UUID or name provided here will be used during '
+                      'automated_cleaning for nodes which do not have a '
+                      'resource_class-specific runbook or runbook set in '
+                      'driver_info.')),
+    cfg.DictOpt('automated_cleaning_runbook_by_resource_class',
+                default={},
+                mutable=True,
+                help=_('A dictionary of key-value pairs of node '
+                       'resource_class and runbook UUID or name which will be '
+                       'used to clean the node if '
+                       '[conductor]automated_clean_step_source is set to '
+                       '\'hybrid\' or \'runbook\' and a more specific runbook '
+                       'has not been configured in driver_info.')),
+    cfg.BoolOpt('automated_cleaning_runbook_from_node',
+                default=False,
+                mutable=True,
+                help=_('When enabled, allows an administrator to configure '
+                       'a runbook in '
+                       'node[\'driver_info\'][\'cleaning_runbook\'] to '
+                       'use for that node when '
+                       '[conductor]automated_clean_step_source is set to '
+                       '\'hybrid\' or \'runbook\'. '
+                       'NOTE: This will permit any user with access to edit '
+                       'node[\'driver_info\'] to circumvent cleaning.')),
+    cfg.BoolOpt('automated_cleaning_runbook_validate_traits',
+                default=True,
+                mutable=True,
+                help=_('When enabled, this option requires validation of a '
+                       'runbook before it\'s used for automated cleaning. '
+                       'Nodes configured with a runbook that is not validated '
+                       'for use via trait matching will fail to clean.')),
     cfg.BoolOpt('allow_provisioning_in_maintenance',
                 default=True,
                 mutable=True,
@@ -279,7 +339,8 @@ opts = [
                       'partition image containing EFI boot loader. This image '
                       'will be used by ironic when building UEFI-bootable ISO '
                       'out of kernel and ramdisk. Required for UEFI boot from '
-                      'partition images.')),
+                      'partition images. Can be overridden per-architecture '
+                      'using the bootloader_by_arch option.')),
     cfg.MultiOpt('clean_step_priority_override',
                  item_type=types.Dict(),
                  default={},
@@ -523,6 +584,15 @@ opts = [
                        'functionality by setting this option to True will '
                        'create a more secure environment, however it may '
                        'break users in an unexpected fashion.')),
+    cfg.BoolOpt('disable_zstandard_decompression',
+                default=False,
+                mutable=False,
+                help=_('Option to enable disabling transparent decompression '
+                       'of files which are compressed with Zstandard '
+                       'compression. This option is provided should operators '
+                       'wish to disable this functionality, otherwise it is '
+                       'automatically applied by the conductor should a '
+                       'compressed artifact be detected.')),
     cfg.ListOpt('file_url_allowed_paths',
                 default=['/var/lib/ironic', '/shared/html', '/templates',
                          '/opt/cache/files', '/vagrant'],
@@ -537,6 +607,67 @@ opts = [
                     'here are validated as absolute paths and will be rejected'
                     'if they contain path traversal mechanisms, such as "..".'
                 )),
+    cfg.IntOpt('graceful_shutdown_timeout',
+               deprecated_group='DEFAULT',
+               deprecated_reason=_(
+                   'This replaces oslo.service '
+                   '[DEFAULT]/graceful_shutdown_timeout option.'),
+               default=60,
+               help='Specify a timeout after which a gracefully shutdown '
+                    'conductor will exit. Zero value means endless wait.'),
+    cfg.DictOpt('bootloader_by_arch',
+                default={},
+                help=_(
+                    'Bootloader ESP image parameter per node architecture. '
+                    'For example: x86_64:bootx64.efi,aarch64:grubaa64.efi. '
+                    'A node\'s cpu_arch property is used as the key to get '
+                    'the appropriate bootloader ESP image. If the node\'s '
+                    'cpu_arch is not in the dictionary, '
+                    'the [conductor]bootloader value will be used instead.'
+                )),
+    cfg.BoolOpt('disable_configdrive_check',
+                default=False,
+                mutable=True,
+                help=_('Option to disable operations which check and '
+                       'potentially fix up configuration drive contents, '
+                       'such as invalid network metadata values. When these '
+                       'issues are detected, and Ironic is able to correct '
+                       'the data, Ironic will do so transparently. Setting '
+                       'this option to True will disable this '
+                       'functionality.')),
+    cfg.BoolOpt('disable_metadata_mtu_check',
+                default=False,
+                mutable=True,
+                help=_('Option to disable consideration of supplied '
+                       'network_data.json link MTU values as basis to '
+                       'regenerate the supplied metadata.')),
+    cfg.BoolOpt('error_on_ramdisk_config_inconsistency',
+                default=False,
+                mutable=True,
+                help=_('Option to determine if Ironic should fail to boot '
+                       'ramdisk in situations where configuration is '
+                       'ambiguous.e.g. if node[driver_info] contains an '
+                       'override for deploy_ramdisk but not deploy_kernel '
+                       'when ambiguous. When set to True, Ironic will raise '
+                       'and fail the provisioning action that required a '
+                       'ramdisk and kernel. When set to False, Ironic will '
+                       'fallback to the next valid, consistent configured '
+                       'ramdisk and kernel for the node.')),
+    cfg.BoolOpt(
+        "record_step_flows_in_history",
+        default=True,
+        help=(
+            "When True, the conductor writes a Node History entry at the "
+            "start and end of every cleaning/servicing/deploy-steps flow. "
+            "Disable this in very high-churn environments to reduce DB load."
+        )),
+    cfg.BoolOpt(
+        'log_step_flows_to_syslog',
+        default=False,
+        help=(
+            "Log steps at the start/end of cleaning/servicing/deployment "
+            "to the conductor service log (WARNING for aborted/failure, "
+            "INFO otherwise.")),
 ]
 
 
