@@ -13,15 +13,21 @@
 # limitations under the License.
 
 from http import client as http_client
+import os
+import ssl
+import time
 
-from ironic_lib import metrics_utils
 from oslo_log import log
 from oslo_serialization import jsonutils
+from oslo_utils import excutils
+from oslo_utils import strutils
 import requests
-import retrying
+import tenacity
 
 from ironic.common import exception
 from ironic.common.i18n import _
+from ironic.common import metrics_utils
+from ironic.common import utils
 from ironic.conf import CONF
 
 LOG = log.getLogger(__name__)
@@ -33,6 +39,15 @@ DEFAULT_IPA_PORTAL_PORT = 3260
 REBOOT_COMMAND = 'run_image'
 
 
+def get_client(task):
+    """Get client for this node."""
+    try:
+        return task.cached_agent_client
+    except AttributeError:
+        task.cached_agent_client = AgentClient()
+        return task.cached_agent_client
+
+
 def get_command_error(command):
     """Extract an error string from the command result.
 
@@ -51,6 +66,15 @@ def get_command_error(command):
         return error
 
 
+def _sanitize_for_logging(var):
+    if not var:
+        return var
+    elif isinstance(var, str):
+        return strutils.mask_password(var)
+    else:
+        return utils.remove_large_keys(strutils.mask_dict_password(var))
+
+
 class AgentClient(object):
     """Client for interacting with nodes via a REST API."""
     @METRICS.timer('AgentClient.__init__')
@@ -77,7 +101,22 @@ class AgentClient(object):
         })
 
     def _get_verify(self, node):
-        return node.driver_info.get('agent_verify_ca', True)
+        # False is a valid value, don't use 'or'
+        value = node.driver_internal_info.get('agent_verify_ca')
+        if value is None:
+            value = node.driver_info.get('agent_verify_ca')
+        if value is None:
+            value = CONF.agent.verify_ca
+
+        if isinstance(value, str):
+            try:
+                value = strutils.bool_from_string(value, strict=True)
+            except ValueError:
+                if not os.path.exists(value):
+                    raise exception.InvalidParameterValue(
+                        _('Agent CA %s is neither a path nor a boolean')
+                        % value)
+        return value
 
     def _raise_if_typeerror(self, result, node, method):
         error = result.get('command_error')
@@ -90,11 +129,12 @@ class AgentClient(object):
                                           error=get_command_error(result))
 
     @METRICS.timer('AgentClient._wait_for_command')
-    @retrying.retry(
-        retry_on_exception=(
-            lambda e: isinstance(e, exception.AgentCommandTimeout)),
-        stop_max_attempt_number=CONF.agent.command_wait_attempts,
-        wait_fixed=CONF.agent.command_wait_interval * 1000)
+    @tenacity.retry(
+        retry=tenacity.retry_if_exception_type(
+            exception.AgentCommandTimeout),
+        stop=tenacity.stop_after_attempt(CONF.agent.command_wait_attempts),
+        wait=tenacity.wait_fixed(CONF.agent.command_wait_interval),
+        reraise=True)
     def _wait_for_command(self, node, method):
         """Wait for a command to complete.
 
@@ -102,23 +142,11 @@ class AgentClient(object):
         :param method: A string represents the command executed by agent.
         :raises: AgentCommandTimeout if timeout is reached.
         """
-        try:
-            method = method.split('.', 1)[1]
-        except IndexError:
-            pass
-
         # NOTE(dtantsur): this function uses AgentCommandTimeout on every
         # failure, but unless the timeout is reached, the exception is caught
         # and retried by the @retry decorator above.
-
-        commands = self.get_commands_status(node)
-        try:
-            result = next(c for c in reversed(commands)
-                          if c.get('command_name') == method)
-        except StopIteration:
-            LOG.debug('Command %(cmd)s is not in the executing commands list '
-                      'for node %(node)s',
-                      {'cmd': method, 'node': node.uuid})
+        result = self.get_last_command_status(node, method)
+        if result is None:
             raise exception.AgentCommandTimeout(command=method, node=node.uuid)
 
         if result.get('command_status') == 'RUNNING':
@@ -133,10 +161,11 @@ class AgentClient(object):
             return result
 
     @METRICS.timer('AgentClient._command')
-    @retrying.retry(
-        retry_on_exception=(
-            lambda e: isinstance(e, exception.AgentConnectionFailed)),
-        stop_max_attempt_number=CONF.agent.max_command_attempts)
+    @tenacity.retry(
+        retry=tenacity.retry_if_exception_type(
+            exception.AgentConnectionFailed),
+        stop=tenacity.stop_after_attempt(CONF.agent.max_command_attempts),
+        reraise=True)
     def _command(self, node, method, params, wait=False, poll=False):
         """Sends command to agent.
 
@@ -152,6 +181,10 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
+        :raises: AgentConnectionFailed when connectivity has failed and there
+                 is no prior command to determine the status of.
         :returns: A dict containing command result from agent, see
                   get_commands_status for a sample.
         """
@@ -165,20 +198,20 @@ class AgentClient(object):
         agent_token = node.driver_internal_info.get('agent_secret_token')
         if agent_token:
             request_params['agent_token'] = agent_token
-        LOG.debug('Executing agent command %(method)s for node %(node)s',
-                  {'node': node.uuid, 'method': method})
+        LOG.debug('Executing agent command %(method)s for node %(node)s '
+                  'with params %(params)s',
+                  {'node': node.uuid, 'method': method,
+                   'params': _sanitize_for_logging(request_params)})
 
         try:
             response = self.session.post(
                 url, params=request_params, data=body,
                 verify=self._get_verify(node),
                 timeout=CONF.agent.command_timeout)
-        except (requests.ConnectionError, requests.Timeout) as e:
-            msg = (_('Failed to connect to the agent running on node %(node)s '
-                     'for invoking command %(method)s. Error: %(error)s') %
-                   {'node': node.uuid, 'method': method, 'error': e})
-            LOG.error(msg)
-            raise exception.AgentConnectionFailed(reason=msg)
+        except (requests.ConnectionError, requests.Timeout, ssl.SSLError) as e:
+            result = self._handle_timeout_on_command_execution(node, method,
+                                                               params, e)
+            response = None
         except requests.RequestException as e:
             msg = (_('Error invoking agent command %(method)s for node '
                      '%(node)s. Error: %(error)s') %
@@ -186,53 +219,56 @@ class AgentClient(object):
             LOG.error(msg)
             raise exception.IronicException(msg)
 
-        # TODO(russellhaering): real error handling
-        try:
-            result = response.json()
-        except ValueError:
-            msg = _(
-                'Unable to decode response as JSON.\n'
-                'Request URL: %(url)s\nRequest body: "%(body)s"\n'
-                'Response status code: %(code)s\n'
-                'Response: "%(response)s"'
-            ) % ({'response': response.text, 'body': body, 'url': url,
-                  'code': response.status_code})
-            LOG.error(msg)
-            raise exception.IronicException(msg)
+        if response is not None:
+            # TODO(russellhaering): real error handling
+            try:
+                result = response.json()
+            except ValueError:
+                msg = _(
+                    'Unable to decode response as JSON.\n'
+                    'Request URL: %(url)s\nRequest body: "%(body)s"\n'
+                    'Response status code: %(code)s\n'
+                    'Response: "%(response)s"'
+                ) % ({'response': response.text, 'body': body, 'url': url,
+                      'code': response.status_code})
+                LOG.error(msg)
+                raise exception.IronicException(msg)
 
         error = result.get('command_error')
         LOG.debug('Agent command %(method)s for node %(node)s returned '
-                  'result %(res)s, error %(error)s, HTTP status code %(code)d',
+                  'result %(res)s, error %(error)s, HTTP status code %(code)s',
                   {'node': node.uuid, 'method': method,
-                   'res': result.get('command_result'),
+                   'res': _sanitize_for_logging(result.get('command_result')),
                    'error': error,
-                   'code': response.status_code})
-        if response.status_code >= http_client.BAD_REQUEST:
+                   'code': response.status_code if response is not None
+                   else 'unknown'})
+        if (response is not None
+                and response.status_code >= http_client.BAD_REQUEST):
             faultstring = result.get('faultstring')
-            if 'agent_token' in faultstring and agent_token:
-                # NOTE(TheJulia) We have an agent that is out of date.
-                # which means I guess grenade updates the agent image
-                # for upgrades... :(
-                if not CONF.require_agent_token:
-                    LOG.warning('Agent command %(method)s for node %(node)s '
-                                'failed. Expected 2xx HTTP status code, got '
-                                '%(code)d. Error suggests an older ramdisk '
-                                'which does not support ``agent_token``. '
-                                'Removing the token for the next retry.',
-                                {'method': method, 'node': node.uuid,
-                                 'code': response.status_code})
-                    i_info = node.driver_internal_info
-                    i_info.pop('agent_secret_token')
-                    node.driver_internal_info = i_info
-                    node.save()
-                    msg = ('Node {} does not appear to support '
-                           'agent_token and it is not required. Next retry '
-                           'will be without the token.').format(node.uuid)
-                    raise exception.AgentConnectionFailed(reason=msg)
-            LOG.error('Agent command %(method)s for node %(node)s failed. '
-                      'Expected 2xx HTTP status code, got %(code)d.',
-                      {'method': method, 'node': node.uuid,
-                       'code': response.status_code})
+            if 'agent_token' in faultstring:
+                LOG.error('Agent command %(method)s for node %(node)s '
+                          'failed. Expected 2xx HTTP status code, got '
+                          '%(code)d. Error suggests an older ramdisk '
+                          'which does not support ``agent_token``. '
+                          'This is a fatal error.',
+                          {'method': method, 'node': node.uuid,
+                           'code': response.status_code})
+            else:
+                LOG.error('Agent command %(method)s for node %(node)s failed. '
+                          'Expected 2xx HTTP status code, got %(code)d.',
+                          {'method': method, 'node': node.uuid,
+                           'code': response.status_code})
+            if (response.status_code == http_client.CONFLICT
+                or 'agent is busy' in faultstring.lower()):
+                # HTTP 409 check as an explicit check of if the agent
+                # is already busy.
+                # NOTE(TheJulia): The agent sends upper case A as of
+                # late victoria, but lower case the entire message
+                # for compatibility with pre-late victoria agents
+                # which returns HTTP 409.
+                raise exception.AgentInProgress(node=node.uuid,
+                                                command=method,
+                                                error=faultstring)
             raise exception.AgentAPIError(node=node.uuid,
                                           status=response.status_code,
                                           error=faultstring)
@@ -282,9 +318,15 @@ class AgentClient(object):
         url = self._get_command_url(node)
         LOG.debug('Fetching status of agent commands for node %s', node.uuid)
 
+        request_params = {}
+        agent_token = node.driver_internal_info.get('agent_secret_token')
+        if agent_token:
+            request_params['agent_token'] = agent_token
+
         def _get():
             try:
                 return self.session.get(url,
+                                        params=request_params,
                                         verify=self._get_verify(node),
                                         timeout=CONF.agent.command_timeout)
             except (requests.ConnectionError, requests.Timeout) as e:
@@ -297,76 +339,101 @@ class AgentClient(object):
                 raise exception.AgentConnectionFailed(reason=msg)
 
         if retry_connection:
-            _get = retrying.retry(
-                retry_on_exception=(
-                    lambda e: isinstance(e, exception.AgentConnectionFailed)),
-                stop_max_attempt_number=CONF.agent.max_command_attempts)(_get)
+            _get = tenacity.retry(
+                retry=tenacity.retry_if_exception_type(
+                    exception.AgentConnectionFailed),
+                stop=tenacity.stop_after_attempt(
+                    CONF.agent.max_command_attempts),
+                reraise=True)(_get)
 
         result = _get().json()['commands']
         status = '; '.join('%(cmd)s: result "%(res)s", error "%(err)s"' %
                            {'cmd': r.get('command_name'),
-                            'res': r.get('command_result'),
+                            'res': _sanitize_for_logging(
+                                r.get('command_result')),
                             'err': r.get('command_error')}
                            for r in result)
         LOG.debug('Status of agent commands for node %(node)s: %(status)s',
                   {'node': node.uuid, 'status': status})
         return result
 
-    @METRICS.timer('AgentClient.prepare_image')
-    def prepare_image(self, node, image_info, wait=False):
-        """Call the `prepare_image` method on the node.
+    def _status_if_last_command_matches(self, node, method, params):
+        """Return the status of the given command if it's the last running."""
+        try:
+            method = method.split('.', 1)[1]
+        except IndexError:
+            pass
+
+        commands = self.get_commands_status(node)
+        if not commands:
+            return None
+
+        # TODO(dtantsur): a more reliable way to detect repeated execution
+        # would be to pass a sort of require ID to the agent.
+
+        command = commands[-1]
+        if command['command_name'] != method:
+            LOG.debug('Command %(cmd)s is not currently executing, the last '
+                      'command is %(curr)s',
+                      {'cmd': method, 'curr': command['command_name']})
+            return None
+
+        if command['command_status'] != 'RUNNING':
+            LOG.debug('Command %(cmd)s is not currently executing, its status '
+                      'is %(curr)s',
+                      {'cmd': method, 'curr': command['command_status']})
+            return None
+
+        return command
+
+    def _handle_timeout_on_command_execution(self, node, method, params,
+                                             error):
+        result = None
+        # NOTE(dtantsur): it is possible, especially with eventlet+TLS, that
+        # agent receives a command but fails to return the result to Ironic.
+        # To avoid a failure, check if the last command is the one we're trying
+        # to execute.
+        try:
+            result = self._status_if_last_command_matches(node, method, params)
+        except Exception as e:
+            msg = (_('Failed to connect to the agent running on node '
+                     '%(node)s for checking the last command status '
+                     'after failing to invoke command %(method)s. '
+                     'Error: %(error)s') %
+                   {'node': node.uuid, 'method': method, 'error': e})
+            LOG.error(msg)
+
+        if result is None:
+            msg = (_('Failed to connect to the agent running on node %(node)s '
+                     'for invoking command %(method)s. Error: %(error)s') %
+                   {'node': node.uuid, 'method': method, 'error': error})
+            LOG.error(msg)
+            raise exception.AgentConnectionFailed(reason=msg)
+
+        return result
+
+    def get_last_command_status(self, node, method):
+        """Get the last status for the given command.
 
         :param node: A Node object.
-        :param image_info: A dictionary containing various image related
-                           information.
-        :param wait: True to wait for the command to finish executing, False
-                     otherwise.
-        :raises: IronicException when failed to issue the request or there was
-                 a malformed response from the agent.
-        :raises: AgentAPIError when agent failed to execute specified command.
-        :returns: A dict containing command status from agent.
-                  See :func:`get_commands_status` for a command result sample.
+        :param method: Command name.
+        :returns: A dict containing command status from agent or None
+            if the command was not found.
         """
-        LOG.debug('Preparing image %(image)s on node %(node)s.',
-                  {'image': image_info.get('id'),
-                   'node': node.uuid})
-        params = {'image_info': image_info}
+        try:
+            method = method.split('.', 1)[1]
+        except IndexError:
+            pass
 
-        # this should be an http(s) URL
-        configdrive = node.instance_info.get('configdrive')
-        if configdrive is not None:
-            params['configdrive'] = configdrive
-
-        return self._command(node=node,
-                             method='standby.prepare_image',
-                             params=params,
-                             poll=wait)
-
-    @METRICS.timer('AgentClient.start_iscsi_target')
-    def start_iscsi_target(self, node, iqn,
-                           portal_port=DEFAULT_IPA_PORTAL_PORT,
-                           wipe_disk_metadata=False):
-        """Expose the node's disk as an ISCSI target.
-
-        :param node: an Ironic node object
-        :param iqn: iSCSI target IQN
-        :param portal_port: iSCSI portal port
-        :param wipe_disk_metadata: True if the agent should wipe first the
-                                   disk magic strings like the partition
-                                   table, RAID or filesystem signature.
-        :raises: IronicException when failed to issue the request or there was
-                 a malformed response from the agent.
-        :raises: AgentAPIError when agent failed to execute specified command.
-        :returns: A dict containing command response from agent.
-                  See :func:`get_commands_status` for a command result sample.
-        """
-        params = {'iqn': iqn,
-                  'portal_port': portal_port,
-                  'wipe_disk_metadata': wipe_disk_metadata}
-        return self._command(node=node,
-                             method='iscsi.start_iscsi_target',
-                             params=params,
-                             wait=True)
+        commands = self.get_commands_status(node)
+        try:
+            return next(c for c in reversed(commands)
+                        if c.get('command_name') == method)
+        except StopIteration:
+            LOG.debug('Command %(cmd)s is not in the executing commands list '
+                      'for node %(node)s',
+                      {'cmd': method, 'node': node.uuid})
+            return None
 
     @METRICS.timer('AgentClient.install_bootloader')
     def install_bootloader(self, node, root_uuid, target_boot_mode,
@@ -387,6 +454,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
                   See :func:`get_commands_status` for a command result sample.
         """
@@ -443,6 +512,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
             See :func:`get_commands_status` for a command result sample.
             The value of key command_result is in the form of:
@@ -474,6 +545,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
             See :func:`get_commands_status` for a command result sample.
             The value of key command_result is in the form of:
@@ -506,6 +579,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentConnectionFailed when an a transient connection failure
+                 breaks the connection while the request is being processed.
         :returns: A dict containing command response from agent.
             See :func:`get_commands_status` for a command result sample.
             The value of key command_result is in the form of:
@@ -537,6 +612,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
             See :func:`get_commands_status` for a command result sample.
             The value of key command_result is in the form of:
@@ -551,7 +628,7 @@ class AgentClient(object):
         """
         params = {
             'step': step,
-            'node': node.as_dict(secure=True),
+            'node': node.as_dict(secure=True, mask_configdrive=False),
             'ports': [port.as_dict() for port in ports],
             'deploy_version': node.driver_internal_info.get(
                 'hardware_manager_version')
@@ -560,6 +637,89 @@ class AgentClient(object):
                              method='deploy.execute_deploy_step',
                              params=params)
 
+    @METRICS.timer('AgentClient.get_service_steps')
+    def get_service_steps(self, node, ports):
+        """Get service steps from agent.
+
+        :param node: A node object.
+        :param ports: Ports associated with the node.
+        :raises: IronicException when failed to issue the request or there was
+                 a malformed response from the agent.
+        :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
+        :returns: A dict containing command response from agent.
+            See :func:`get_commands_status` for a command result sample.
+            The value of key command_result is in the form of:
+
+            ::
+
+              {
+                'service_steps': <a list of service steps>,
+                'hardware_manager_version': <manager version>
+              }
+
+        """
+        params = {
+            'node': node.as_dict(secure=True),
+            'ports': [port.as_dict() for port in ports]
+        }
+        try:
+            response = self._command(node=node,
+                                     method='service.get_service_steps',
+                                     params=params,
+                                     wait=True)
+        except exception.AgentAPIError:
+            # NOTE(TheJulia): This seems logical to do to handle an
+            # older ironic-python-agent, since the net-effect will be
+            # "there is nothing we can issue to the agent".
+            # TODO(TheJulia): Once we know the version where this *is*
+            # supported, we should actually update this log message.
+            # We won't know that until after the initial merge.
+            LOG.warning('Unable to retrieve service steps for node %s.'
+                        'Please upgrade your ironic-python-agent.',
+                        node.uuid)
+            response = {
+                'service_steps': [],
+                'hardware_manager_version': 0,
+            }
+        return response
+
+    @METRICS.timer('AgentClient.execute_service_step')
+    def execute_service_step(self, step, node, ports):
+        """Execute specified service step.
+
+        :param step: A service step dictionary to execute.
+        :param node: A Node object.
+        :param ports: Ports associated with the node.
+        :raises: IronicException when failed to issue the request or there was
+                 a malformed response from the agent.
+        :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
+        :returns: A dict containing command response from agent.
+            See :func:`get_commands_status` for a command result sample.
+            The value of key command_result is in the form of:
+
+            ::
+
+              {
+                'service_result': <the result of execution, step specific>,
+                'service_step': <the service step issued to agent>
+              }
+
+        """
+        params = {
+            'step': step,
+            'node': node.as_dict(secure=True),
+            'ports': [port.as_dict() for port in ports],
+            'service_version': node.driver_internal_info.get(
+                'hardware_manager_version')
+        }
+        return self._command(node=node,
+                             method='service.execute_service_step',
+                             params=params)
+
     @METRICS.timer('AgentClient.get_partition_uuids')
     def get_partition_uuids(self, node):
         """Get deploy steps from agent.
@@ -568,6 +728,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
 
         """
@@ -584,6 +746,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
                   See :func:`get_commands_status` for a command result sample.
         """
@@ -599,6 +763,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
                   See :func:`get_commands_status` for a command result sample.
         """
@@ -614,6 +780,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
                   See :func:`get_commands_status` for a command result sample.
         """
@@ -630,6 +798,8 @@ class AgentClient(object):
         :raises: IronicException when failed to issue the request or there was
                  a malformed response from the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :returns: A dict containing command response from agent.
                   See :func:`get_commands_status` for a command result sample.
         """
@@ -647,6 +817,8 @@ class AgentClient(object):
                  to issue the request, or there was a malformed response from
                  the agent.
         :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
         :raises: InstanceRescueFailure when the agent ramdisk is too old
                  to support transmission of the rescue password.
         :returns: A dict containing command response from agent.
@@ -679,3 +851,75 @@ class AgentClient(object):
                 return self._command(node=node,
                                      method='rescue.finalize_rescue',
                                      params=params)
+
+    @METRICS.timer('AgentClient.lockdown')
+    def lockdown(self, node, fail_if_unavailable=True):
+        """Lock down the agent so that it's not usable any more.
+
+        :param node: A Node object.
+        :param fail_if_unavailable: Whether to fail this call if the agent is
+                                    already unavailable.
+        :raises: IronicException when failed to issue the request or there was
+                 a malformed response from the agent.
+        :raises: AgentAPIError when agent failed to execute specified command.
+        :raises: AgentInProgress when the command fails to execute as the agent
+                 is presently executing the prior command.
+        """
+        if not fail_if_unavailable:
+            try:
+                self.get_commands_status(node, expect_errors=True)
+            except exception.AgentConnectionFailed:
+                LOG.debug('Agent is already down when trying to lock down '
+                          'node %s', node.uuid)
+                return
+            except Exception:
+                LOG.exception('Unexpected exception when checking agent '
+                              'status on node %s, proceeding with lockdown',
+                              node.uuid)
+
+        wait = CONF.agent.post_deploy_get_power_state_retry_interval
+        attempts = CONF.agent.post_deploy_get_power_state_retries + 1
+
+        @tenacity.retry(stop=tenacity.stop_after_attempt(attempts),
+                        retry=tenacity.retry_unless_exception_type(
+                            exception.AgentConnectionFailed),
+                        wait=tenacity.wait_fixed(wait),
+                        sleep=time.sleep,  # for unit testing
+                        reraise=True)
+        def _wait_until_locked_down(node):
+            self.get_commands_status(node, expect_errors=True)
+            LOG.debug('Agent is still available on node %s, waiting for the '
+                      'lockdown command to take effect', node.uuid)
+
+        self.sync(node)
+
+        try:
+            self._command(node=node, method='system.lockdown', params={})
+        except Exception as e:
+            with excutils.save_and_reraise_exception():
+                LOG.error('Failed to lock down node %(node_uuid)s. '
+                          '%(cls)s: %(error)s',
+                          {'node_uuid': node.uuid,
+                           'cls': e.__class__.__name__, 'error': e},
+                          exc_info=not isinstance(
+                              e, exception.IronicException))
+
+        try:
+            _wait_until_locked_down(node)
+        except exception.AgentConnectionFailed:
+            pass  # expected
+        except tenacity.RetryError:
+            LOG.error('Failed to lock down node %(node_uuid)s in at least '
+                      '%(timeout)d seconds: agent is still available',
+                      {'node_uuid': node.uuid,
+                       'timeout': (wait * (attempts - 1))})
+            raise exception.AgentCommandTimeout(command='system.lockdown',
+                                                node=node.uuid)
+        except Exception as e:
+            with excutils.save_and_reraise_exception():
+                LOG.error('Failed to lock down node %(node_uuid)s. '
+                          '%(cls)s: %(error)s',
+                          {'node_uuid': node.uuid,
+                           'cls': e.__class__.__name__, 'error': e},
+                          exc_info=not isinstance(
+                              e, exception.IronicException))
